/*
 * Copyright (c) 2012 Citrix Systems, Inc.
 * Copyright (C) 2015 Assured Information Security, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "qemu/osdep.h"
#include "block/block.h"
#include "sysemu/blockdev.h"
#include "sysemu/block-backend.h"
#include "qemu/timer.h"
#include "hw/xen/xen_backend.h"
#include "qmp-commands.h"
#include "xen-changeiso.h"

static QTAILQ_HEAD(IsoNodeHead, IsoNode) iso_list = QTAILQ_HEAD_INITIALIZER(iso_list);

static void update_iso_cb(void *opaque)
{
    struct IsoNode *node = opaque;
    Error *err = NULL;

    if (!node) {
        return;
    }

    // Don't try to do  the change if node->iso_file doesn't exist
    if (access(node->iso_file, F_OK) != -1) {
        qmp_blockdev_change_medium(blk_name(node->blk),
                                  node->iso_file, false, NULL, false, 0, &err);
    }

    if (err) {
        fprintf(stderr, "%s Error: %s", __FUNCTION__, error_get_pretty(err));
    }
}

// This is the xenstore watch callback.
// What we want to do here is pretty different wether we're in a stubdom or not.
// In the non-stubdom case, we're just watching the xenstore key "params",
//   which contains the path to the current iso. An empty value means eject.
// In the stubdom case, we also watch the "params" key, but it can only contain
//   one of two values: "/dev/xvdc" or "". "" (empty value) means eject.
//   When the iso changes in dom0, param will switch from "/dev/xvdc" to "" and
//   back to "/dev/xvdc". We want to reinsert the CD on that last event, but only
//   once the stubdom is successfully connected to it (state == 4).
//   So we don't only watch params but also state.
static void xenstore_update_iso(void *opaque)
{
#ifdef CONFIG_STUBDOM
    char *state = NULL;
#endif
    char *newFile = NULL;
    struct IsoNode *node = opaque;
    size_t len;
    char path[XEN_BUFSIZE];
    Error *err = NULL;

    if(!xenstore || !node)
        return;

    if(!node->first_watch) {
        memset(path, 0x00, XEN_BUFSIZE);

        // Read the new params, eject if params == ""
        snprintf(path, XEN_BUFSIZE, "%s/params", node->xen_vbd_id);
        newFile = xs_read(xenstore, XBT_NULL, path, &len);

        if (newFile != NULL && *newFile == '\0')
            qmp_eject(blk_name(node->blk), true, true, &err);

#ifdef CONFIG_STUBDOM
        snprintf(path, XEN_BUFSIZE, "%s/state", node->xen_vbd_id);
        state = xs_read(xenstore, XBT_NULL, path, &len);

        if (!state)
            return;
#else
        // Switch out the file path to the iso, so change gets made
        // when the node->timer fires.
        if (newFile) {
            g_free(node->iso_file);
            node->iso_file = NULL;
            node->iso_file = g_strdup(newFile);

            if (!node->iso_file) {
                free(newFile);
                return;
            }
        }
#endif

#ifdef CONFIG_STUBDOM
        // Wait for blkfront<->blkback to get ready before changing
        // the disk
        if (!strcmp(state, BLKFRONT_READY_STATE)) {
            if (node->just_started) {
                // blkfront just finished setting up /dev/xvdc
                // doing an eject and re-insert at this point upsets seabios
                // let's do nothing
                node->just_started = false;
            } else {
#endif
                // If newFile is "", we already ejected.
                if (newFile && *newFile != '\0')
                    timer_mod(node->timer, qemu_clock_get_ms(QEMU_CLOCK_REALTIME) + 1000);
#ifdef CONFIG_STUBDOM
            }
        }

        if(state) {
            free(state);
            state = NULL;
        }
#endif
        if(newFile) {
            free(newFile);
            newFile = NULL;
        }
    } else {
        node->first_watch = false;
    }
}

/*

*/
int xenstore_register_iso_dev(const char *file, BlockBackend *blk)
{
    struct IsoNode *node = NULL;

    if (!file || !blk) {
        fprintf(stderr, "Failed to register iso device due to incorrect parameters");
        return -EINVAL;
    }

    node = g_malloc0(sizeof(*node));

    if (!node) {
        return -ENOMEM;
    }

    /*
       We can't do our Xen business just yet, because hvm init domain hasn't
       been called yet. So we have to save any information needed for registering
       until later.
    */

    node->iso_file = g_strdup(file);
    node->blk = blk;
    node->xen_vbd_id = NULL;
    node->frontend_state = NULL;
    node->just_started = true;
    node->first_watch = true;
    node->timer = timer_new_ms(QEMU_CLOCK_REALTIME, update_iso_cb, node);

    if (!node->iso_file || !node->timer) {
        g_free(node);
        return -ENOMEM;
    }

    QTAILQ_INSERT_TAIL(&iso_list, node, next);

    return 0;
}

int xenstore_init_iso_dev(void)
{
    unsigned int dirNum = 0, i = 0;
    char *dompath = NULL;
    char **vbd_devs = NULL;
    char path[XEN_BUFSIZE];
    char token[XEN_BUFSIZE];

    memset(path, 0x00, XEN_BUFSIZE);
    memset(token, 0x00, XEN_BUFSIZE);

    if (0 > xenstore_generic_init()) {
        return -ENOENT;
    }

    // Stubdom domid is xen_domid+1
#ifdef CONFIG_STUBDOM
    dompath = xs_get_domain_path(xenstore, STUBDOMID(xen_domid));
#else
    dompath = xs_get_domain_path(xenstore, xen_domid);
#endif

    if (!dompath) {
        fprintf(stderr, "%s: Failed to retrieve dompath", __FUNCTION__);
        return -1;
    }

    snprintf(path, XEN_BUFSIZE, "%s/device/vbd", dompath);

    // Find the virtual-device id that blkfront is using for this device
    vbd_devs = xs_directory(xenstore, XBT_NULL, path, &dirNum);

    if (!vbd_devs) {
        return -1;
    }

    for (i = 0; i < dirNum; i++) {

        if (!vbd_devs[i]) {
            continue;
        }

        // Build paths to get necessary information from Xenstore
        // Check the device type as CDROM, and get the backend path
        memset(path, 0x00, XEN_BUFSIZE);
        snprintf(path, XEN_BUFSIZE, "%s/device/vbd/%s", dompath, vbd_devs[i]);
        char *dev_type = xenstore_read_str(path, "device-type");
        char *be = xenstore_read_str(path, "backend");

        if (dev_type && be && !strcmp(dev_type, "cdrom")) {
            // We need to watch the backend for this device now.
            char *params = xenstore_read_str(be, "params");
            struct IsoNode *node;

            QTAILQ_FOREACH(node, &iso_list, next) {
                if (node && params &&
                    node->xen_vbd_id == NULL &&
                    node->frontend_state == NULL) {
                    // For mapping a fired watch to a specific device later
                    node->xen_vbd_id = strdup(be);

                    // Before the guest disk change can occur, make sure the state
                    // of the specified blkfront device is ready
                    memset(path, 0x00, XEN_BUFSIZE);
                    snprintf(path, XEN_BUFSIZE, "%s/device/vbd/%s/state", dompath, vbd_devs[i]);
                    node->frontend_state = strdup(path);
                    break;
                }
            }

            if (xenstore_add_watch(be, "params", xenstore_update_iso, (void *) node) != 0) {
                fprintf(stderr, "[OXT-ISO] Failed to install xenstore watch on path: %s/params", be);
            }

#ifdef CONFIG_STUBDOM
            if (xenstore_add_watch(be, "state", xenstore_update_iso, (void *) node) != 0) {
                fprintf(stderr, "[OXT-ISO] Failed to install xenstore watch on path: %s/state", be);
            }
#endif

            if (be) {
                free(be);
                be = NULL;
            }

            if (dev_type) {
                free(dev_type);
                dev_type = NULL;
            }

            if (params) {
                free(params);
                params = NULL;
            }
        }
    }

    if (dompath) {
        free(dompath);
        dompath = NULL;
    }

    if (vbd_devs) {
        free(vbd_devs);
        vbd_devs = NULL;
    }

    return 0;
}
